/* absdisk.c -- absolute disk access
   Copyright (C) 2016 Joe Forster/STA

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#include "lfnsrv.h"
#include <devioctl.h>
#include <ntdddisk.h>

typedef struct {
  char name[50 + 1];
  HANDLE hDisk;
  WORD wMode;
  DISK_GEOMETRY diskgeo;
} BIOSDISK, *PBIOSDISK;

typedef struct {
  BYTE size;
  BYTE reserved;
  WORD wNumSec;
  ULONG buffer;
  DWORD64 qwStartSec;
  DWORD64 bufferLinear;
} BIOSDISKXFERPACKET, *PBIOSDISKXFERPACKET;

WORD open_disk(
		IN OUT PBIOSDISK disk,
		IN  BYTE nDrive,
		IN  BOOL bPhysical,
		IN  WORD wMode)
{
	char*	name;
	HANDLE	hDisk;
	DWORD	dummy;
	ISNULL(disk);

	disk->hDisk = INVALID_HANDLE_VALUE;
	disk->wMode = 0;
	strcpy(disk->name, "\\\\.\\");
	name = &disk->name[4];
	if (bPhysical) {
		if (nDrive < 0x80) {
			strcpy(name, "GLOBALROOT\\Device\\Floppy0");
		} else {
			strcpy(name, "PhysicalDrive0");
			nDrive -= 0x80;
		}
		while (*name) name++;
                *(name - 1) = nDrive + '0';
		name = disk->name;
	} else {
		if (nDrive > 'Z' - 'A') {
			return -1;
		}
		if (nDrive > 0) {
			strcpy(name, "A:");
			*name = nDrive + ('A' - 1);
			name = disk->name;
		} else {
			*disk->name = '\0';
			name = NULL;
		}
	}
	disk->hDisk = CreateFile(name, (wMode & 1) ? GENERIC_WRITE : GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);
	if (disk->hDisk != INVALID_HANDLE_VALUE) {
		if (!DeviceIoControl(disk->hDisk, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &disk->diskgeo, sizeof(disk->diskgeo), &dummy, NULL)) {
			CloseHandle(disk->hDisk);
			DBGMSG("DeviceIoControl call failed");
			disk->hDisk = INVALID_HANDLE_VALUE;
			return -1;
		}
	} else {
		DBGMSG("CreateFile call failed");
		return -1;
	}
	disk->wMode = wMode;
	return 0;
}

WORD	seek_disk(
		IN  PBIOSDISK disk,
		IN  DWORD64 qwSector)
{
	DWORD64	qwPosition = qwSector * disk->diskgeo.BytesPerSector;
	ISNULL(disk);

	if (SetFilePointer(disk->hDisk, (DWORD)qwPosition, (PDWORD)&qwPosition + 1, FILE_BEGIN) == -1) {
		DBGMSG("SetFilePointer call failed");
		return -1;
	}
	return 0;
}

typedef BOOL (*PFN_READ_WRITE_FILE)(HANDLE, LPVOID, DWORD, LPDWORD, LPOVERLAPPED);

WORD	transfer_disk(
		IN  PBIOSDISK disk,
		IN  DWORD dwNumSec,
		IN  OUT PDWORD dwNumSecTransferred,
		IN  ULONG BufferAddress)
{
        WORD	result = 0;
	DWORD	dwNumBytes = (DWORD)dwNumSec * disk->diskgeo.BytesPerSector;
	DWORD	dwBytesRead;
	PVOID	buffer = GetVDMPointer(BufferAddress, dwNumBytes, FALSE);
	PFN_READ_WRITE_FILE xferfunc = (disk->wMode & 1) ? (PFN_READ_WRITE_FILE)WriteFile : ReadFile;
	ISNULL(disk);

	if (xferfunc(disk->hDisk, buffer, dwNumBytes, &dwBytesRead, NULL) == 0) {
			DBGMSG("ReadFile call failed");
			result = -1;
	}
	FreeVDMPointer(BufferAddress,dwNumBytes, buffer, FALSE);
	if (dwNumSecTransferred)
	        *dwNumSecTransferred = dwBytesRead / disk->diskgeo.BytesPerSector;
	if (result == 0 && dwBytesRead < dwNumBytes) {
		result = ERROR_HANDLE_EOF;
	}
	return result;
}

WORD close_disk(
		IN  PBIOSDISK disk)
{
        BOOL result;
	ISNULL(disk);

	result = CloseHandle(disk->hDisk);
	disk->hDisk = INVALID_HANDLE_VALUE;
	return result ? 0 : -1;
}

WORD	process_abs_disk_access(
		IN  BYTE nDrive,
		IN  WORD wMode,
		IN  DWORD dwStartSec,
		IN  WORD wNumSec,
		IN  ULONG BufferAddress)
{
	WORD	result = 0;
	BIOSDISK disk;

	PROLOG(process_abs_disk_access);
	DBGVALUE(nDrive,"%d");
	DBGVALUE(wMode,"%x");
	DBGVALUE(dwStartSec,"%lx");
	DBGVALUE(wNumSec,"%x");

	result = open_disk(&disk, nDrive, FALSE, wMode);
	if (result == 0) {
		result = seek_disk(&disk, dwStartSec);
		if (result == 0)
			result = transfer_disk(&disk, wNumSec, NULL, BufferAddress);
		close_disk(&disk);
	}
	if (result != 0) {
		if (result == -1)
			result = (WORD)GetLastError();
		if (result == ERROR_FILE_NOT_FOUND)
			result = ERROR_INVALID_DRIVE;
		DBGVALUE(result,"%d");
		EPILOG(process_abs_disk_access, FALSE);
		return result;
	}

	EPILOG(process_abs_disk_access,TRUE);
	return ERROR_SUCCESS;
}

WORD	process_bios_disk(void) {
	WORD  result = 0;
	WORD  regAX = HIWORD(getEAX());
	WORD  regAH = HIBYTE(regAX);
	// restore original AL
	setAL(LOBYTE(regAX));
      	switch (regAH) {
//		case BDSK_RESET_DISKS     :
//		case BDSK_RESET_HARD_DISKS: {
//			// do nothing
//			break;
//		}
//		case BDSK_GET_LAST_STATUS: {
//			// do nothing
//			break;
//		}
		case BDSK_CHECK_EXTENSIONS: {
			WORD BX = getBX();
			if (BX == BDSK_CHECK_EXTENSIONS_MAGIC) {
				setBX(BX ^ (WORD)-1);
				// EDD 3.0, all functions supported
				setAH(0x30);
				setCX(0x0007);
				setDH(0);
			} else {
				result = ERROR_INVALID_FUNCTION;
			}
			break;
		}
		default: {
		        BYTE nDrive = getDL();
		        WORD wMode = (regAH == BDSK_WRITE_SECTORS || regAH == BDSK_EXT_WRITE_SECTORS) ? 1 : 0;
		        DWORD dummy;
			BIOSDISK disk;
			result = open_disk(&disk, nDrive, TRUE, wMode);
			if (result == 0) {
				switch (regAH) {
		case BDSK_READ_SECTORS      :
		case BDSK_WRITE_SECTORS     :
		case BDSK_VERIFY_SECTORS    :
		case BDSK_EXT_READ_SECTORS  :
		case BDSK_EXT_WRITE_SECTORS :
		case BDSK_EXT_VERIFY_SECTORS:
		case BDSK_EXT_SEEK          : {
			BOOL extended = (regAH >= BDSK_EXT_READ_SECTORS);
			DWORD64 qwStartSec;
			DWORD dwNumSec;
			DWORD dwNumSecTransferred;
			ULONG PacketAddress;
			PBIOSDISKXFERPACKET packet = NULL;
			ULONG BufferAddress;
			if (extended) {
			        PacketAddress = MAKEADDRESS(getDS(), getSI());
				packet = (PBIOSDISKXFERPACKET) GetVDMPointer(PacketAddress, sizeof(BIOSDISKXFERPACKET), FALSE);
				if (packet->size == sizeof(BIOSDISKXFERPACKET) || packet->size == sizeof(BIOSDISKXFERPACKET) - sizeof(DWORD64)) {
					qwStartSec = packet->qwStartSec;
					dwNumSec = packet->wNumSec;
					BufferAddress = packet->buffer;
				} else {
					result = ERROR_INVALID_PARAMETER;
				}
			} else {
				WORD cylinder = (((WORD)getCL() & 0xC0) << 2) | getCH();
				BYTE head     = getDH();
				BYTE sector   = getCL() & 0x3F;
				if (sector != 0) {
					qwStartSec = ((DWORD)cylinder * disk.diskgeo.TracksPerCylinder + head) * disk.diskgeo.SectorsPerTrack + (sector & 0x3F) - 1;
					dwNumSec = LOBYTE(regAX);
					BufferAddress = MAKEADDRESS(getES(), getBX());
				} else {
					result = ERROR_INVALID_PARAMETER;
				}
			}
			if (result == 0)
				result = seek_disk(&disk, qwStartSec);
			if (result == 0 && regAH != BDSK_EXT_SEEK) {
				result = transfer_disk(&disk, dwNumSec, &dwNumSecTransferred, BufferAddress);
				if (result == 0)
					if (extended) {
						packet->wNumSec = LOWORD(dwNumSecTransferred);
					} else
						setAL(LOBYTE(dwNumSecTransferred));
			}
			if (packet != NULL)
				FreeVDMPointer(PacketAddress,sizeof(BIOSDISKXFERPACKET), packet, FALSE);
			break;
		}
//		case BDSK_FORMAT_TRACK: {
//			// do nothing
//			break;
//		}
//		case BDSK_GET_DRIVE_PARAMS:
		case BDSK_EXT_GET_DRIVE_PARAMS: {
			// do nothing
			break;
		}
//		case BDSK_CHECK_READY: {
//			if (!DeviceIoControl(disk.hDisk, IOCTL_STORAGE_CHECK_VERIFY, NULL, 0, NULL, 0, &dummy, NULL))
//				result = -1;
//			break;
//		}
//		case BDSK_GET_DISK_TYPE: {
//			if (disk.diskgeo.MediaType == FixedMedia) {
//				DWORD64 totalSec = (DWORD64)disk.diskgeo.Cylinders.LowPart * disk.diskgeo.TracksPerCylinder * disk.diskgeo.SectorsPerTrack;
//				setAL(0x03);
//				setDX(LOWORD(totalSec));
//				setCX(HIWORD(totalSec));
//			} else
//				setAL(0x02);
//			break;
//		}
//		case BDSK_DETECT_DISK_CHANGE:
		case BDSK_EXT_DETECT_MEDIA_CHANGE: {
			// do nothing
			break;
		}
		case BDSK_EXT_LOCK_DRIVE:
		case BDSK_EXT_EJECT_MEDIA: {
			// do nothing
			break;
		}
		default:
			result = ERROR_INVALID_FUNCTION;
			break;
				}
				close_disk(&disk);
			}
			if (result != 0) {
				if (result == (WORD)-1)
					result = (WORD)GetLastError();
				if (result == ERROR_FILE_NOT_FOUND)
					result = ERROR_INVALID_DRIVE;
				}
			break;
		}
	}
	// return error code in AH
	if (result) {
		setCF(1);
		setAH(LOBYTE(result));
	} else {
		setCF(0);
		setAH(0);
	}
	return result;
}

WORD	process_lock_disk(
		IN  BYTE nDrive,
		IN  BYTE bCategory,
		IN  BYTE bMinor,
		IN  PWORD wLockLevel,
		IN  PWORD wPermissions)
{
	WORD	result;

	PROLOG(process_lock_disk);
	DBGVALUE(nDrive,"%d");
	DBGVALUE(bCategory,"%x");
	DBGVALUE(bMinor,"%x");
	DBGVALUE(*bLockLevel,"%x");
	DBGVALUE(*wPermissions,"%x");

	switch (bCategory) {
		case W95_LOCK_DRIVE_CATEG_DISK	 :
		case W95_LOCK_DRIVE_CATEG_FAT32_DISK: {
			switch (bMinor) {
				case W95_LOCK_DRIVE_MINOR_LOCK_LOGICAL   :
				case W95_LOCK_DRIVE_MINOR_LOCK_PHYSICAL  :
				case W95_LOCK_DRIVE_MINOR_UNLOCK_LOGICAL :
				case W95_LOCK_DRIVE_MINOR_UNLOCK_PHYSICAL: {
					// do nothing
					break;
				}
				case W95_LOCK_DRIVE_MINOR_LOCK_GETFLAGS: {
					// nothing happenned
					*wPermissions = 0;
					break;
				}
				case W95_LOCK_DRIVE_MINOR_LOCK_GETSTATE: {
					// disk unlocked
					*wLockLevel = (WORD)-1;
					*wPermissions = 0;
					break;
				}
				default:
					result = ERROR_INVALID_PARAMETER;
					break;
			}
			break;
		}
		default:
			result = ERROR_INVALID_PARAMETER;
			break;
	}

	EPILOG(process_lock_disk,TRUE);
	return ERROR_SUCCESS;
}
